"""This file is a wrapper for Model trainer base class."""

import torch
from torch import nn
from torch.nn import Module
from torch.optim import Optimizer
from torch.utils.data import DataLoader

from scalr.nn.callbacks import CallbackExecutor
from scalr.nn.trainer import TrainerBase


class SimpleModelTrainer(TrainerBase):
    """Class for Simple model trainer.
    
    It works with dataloaders which contain all input tensors in line
    with model input, and the last tensor as target to train the model.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



"""This file is an implementation of the Absolute mean feature selector strategy."""

from pandas import DataFrame

from scalr.feature.selector import SelectorBase


class AbsMean(SelectorBase):
    """Class for absolute mean feature selector strategy.
    
    It uses the absolute mean across all classes as the score of the feature.
    """

    def __init__(self, k: int = 1e6):
        """Initialize required parameters for the selector."""
        self.k = k

    def get_feature_list(self, score_matrix: DataFrame) -> list[str]:
        """A function to return top features using score matrix and selector strategy.

        Args:
            score_matrix (DataFrame): Score of each feature across all classes
                                      [num_classes X num_features].

        Returns:
            list[str]: List of top k features.
        """
        top_features_list = list(score_matrix.abs().mean().sort_values(
            ascending=False).reset_index()['index'][:self.k])
        return top_features_list

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for preprocess_config."""
        return dict(k=int(1e6))



"""This file is an implementation of model checkpoint callback."""

import os
from os import path

import torch

from scalr.nn.callbacks import CallbackBase


class ModelCheckpoint(CallbackBase):
    """Model checkpointing to save model weights at regular intervals.

    Attributes:
        epoch: An interger count of epochs trained.
        max_validation_acc: Keeps track of the maximum validation accuracy across all epochs.
        interval: Regular interval of model checkpointing.
    """

    def __init__(self, dirpath: str, interval: int = 5):
        """Intialize required parameters for model checkpoint callback.

        Args:
            dirpath: To store the respective model checkpoints.
            interval: Regular interval of model checkpointing.
        """

        self.epoch = 0
        self.interval = int(interval)
        self.dirpath = dirpath

        if self.interval:
            os.makedirs(path.join(dirpath, 'checkpoints'), exist_ok=True)

    def save_checkpoint(self, model_state_dict: dict, opt_state_dict: dict,
                        path: str):
        """A function to save model & optimizer state dict to the given path.
        
        Args:
            model_state_dict: Model's state dict.
            opt_state_dict: Optimizer's state dict.
            path: Path to store checkpoint to.
        """
        torch.save(
            {
                'epoch': self.epoch,
                'model_state_dict': model_state_dict,
                'optimizer_state_dict': opt_state_dict
            }, path)

    def __call__(self, model_state_dict: dict, opt_state_dict: dict, **kwargs):
        """A function that evaluates when to save a checkpoint.

        Args:
            model_state_dict: Model's state dict.
            opt_state_dict: Optimizer's state dict.
        """
        self.epoch += 1
        if self.interval and self.epoch % self.interval == 0:
            self.save_checkpoint(
                model_state_dict, opt_state_dict,
                path.join(self.dirpath, 'checkpoints',
                          f'model_{self.epoch}.pt'))

    @classmethod
    def get_default_params(cls):
        """Class method to get default params for model_config."""
        return dict(dirpath='.', interval=5)



"""This file is a base class for loss functions."""

from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
import torch
from torch import nn

import scalr


class CustomLossBase(nn.Module):
    """Base class to implement custom loss functions."""

    def __init__(self):
        super().__init__()
        self.criterion = None

    def forward(self, out, preds):
        """Returns loss betwen outputs and predictions."""
        return self.criterion(out, preds)


def build_loss_fn(loss_config):
    """Builder object to get Loss function, updated loss_config."""
    name = loss_config.get('name')
    if not name:
        raise ValueError('Loss function not provided')

    params = loss_config.get('params', dict())

    # TODO: Add provision for custom loss object
    loss_fn = getattr(torch.nn, name)(**params)
    return loss_fn, loss_config



"""This file returns top K features(can be promoters or inhibitors as well) per class."""

from pandas import DataFrame

from scalr.feature.selector import SelectorBase


class ClasswiseAbs(SelectorBase):
    """Class for class-wise absolute feature selector strategy.
    
    Classwise scorer returns a dict for each class, containing the top
    absolute scores of genes.
    """

    def __init__(self, k: int = 1e6) -> dict:
        """Initialize required parameters for the selector."""
        self.k = k

    def get_feature_list(self, score_matrix: DataFrame):
        """A function to return top features per class using score matrix
        and selector strategy.

        Args:
            score_matrix (DataFrame): Score of each feature across all classes
                                      [num_classes X num_features].

        Returns:
            dict: List of top_k features for each class.
        """
        classwise_abs = dict()
        n_cls = len(score_matrix)

        for i in range(n_cls):
            for i in range(n_cls):
                classwise_abs[score_matrix.index[i]] = abs(
                    score_matrix.iloc[i, :]).sort_values(
                        ascending=False).reset_index()['index'][:self.k].tolist(
                        )

        return classwise_abs

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for preprocess_config."""
        return dict(k=int(1e6))



"""This file generates ROC-AUC plot and stores it."""

from os import path

import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import auc
from sklearn.metrics import roc_curve
from sklearn.metrics import RocCurveDisplay

from scalr.analysis import AnalysisBase
from scalr.utils import data_utils
from scalr.utils import EventLogger


class RocAucCurve(AnalysisBase):
    '''Class to generate ROC-AUC curve.'''

    def generate_analysis(self, test_labels: list[int],
                          pred_probabilities: list[list[float]], dirpath: str,
                          mapping: list, **kwargs) -> None:
        """A function to calculate ROC-AUC and save the plot.

        Args:
            test_labels: True labels from the test dataset.
            pred_probabilities: Predictions probabilities of each sample for all the classes.
            dirpath: Path to store gene recall curve if applicable.
            mapping: List of class names.
        """

        logger_name = "ROC-AUC Analysis"
        self.event_logger = EventLogger(logger_name)
        self.event_logger.heading2(logger_name)
        self.event_logger.info("Generating one hot matrix of test labels.")

        # convert label predictions list to the one-hot matrix.
        test_labels_onehot = data_utils.get_one_hot_matrix(
            np.array(test_labels))
        fig, ax = plt.subplots(1, 1, figsize=(16, 8))

        self.event_logger.info(
            "Calculating ROC-AUC for each label and creating a plot for that.")

        # test labels start with 0 so we need to add 1 in max.
        for class_label in range(max(test_labels) + 1):

            # fpr: False Positive Rate | tpr: True Positive Rate
            fpr, tpr, _ = roc_curve(
                test_labels_onehot[:, class_label],
                np.array(pred_probabilities)[:, class_label])

            roc_auc = auc(fpr, tpr)

            display = RocCurveDisplay(fpr=fpr, tpr=tpr, roc_auc=roc_auc)
            display.plot(ax=ax, name=mapping[class_label])

        self.event_logger.info("Saving plot and clear axis & figure.")

        plt.axline((0, 0), (1, 1), linestyle='--', color='black')
        fig.savefig(path.join(dirpath, f'roc_auc.svg'))
        plt.clf()    # clear axis & figure so it does not affect the next plot.



"""This file is an implementation of early stopping callback."""

import os
from os import path

import torch

from scalr.nn.callbacks import CallbackBase


class EarlyStopping(CallbackBase):
    """
    Implements early stopping based upon validation loss.

    Attributes:
        patience: Number of epochs with no improvement after which training will be stopped.
        min_delta: Minimum change in the monitored quantity to qualify as an improvement,
        i.e. an absolute change of less than min_delta, will count as no improvement.
    """

    def __init__(self,
                 dirpath: str = None,
                 patience: int = 3,
                 min_delta: float = 1e-4):
        """Intialize required parameters for early stopping callback.

        Args:
            patience: Number of epochs with no improvement after which training will be stopped.
            min_delta: Minimum change in the monitored quantity to qualify as an improvement,
                            i.e. an absolute change of less than min_delta, will count as no improvement.
            epoch: An interger count of epochs trained.
            min_validation_loss: Keeps track of the minimum validation loss across all epochs.
        """
        self.patience = int(patience)
        self.min_delta = float(min_delta)
        self.epoch = 0
        self.min_val_loss = float('inf')

    def __call__(self, val_loss: float, **kwargs) -> bool:
        """Return `True` if model training needs to be stopped based upon improvement conditions.
        Else returns `False` for continued training.
        """
        if val_loss < self.min_val_loss:
            self.min_val_loss = val_loss
            self.epoch = 0
        elif val_loss >= (self.min_val_loss + self.min_delta):
            self.epoch += 1
            if self.epoch >= self.patience:
                return True
        return False

    @classmethod
    def get_default_params(cls):
        """Class method to get default params for model_config."""
        return dict(patience=3, min_delta=1e-4)



"""This file is a base class for the top feature selector."""

from typing import Union

from pandas import DataFrame

import scalr
from scalr.utils import build_object


class SelectorBase:
    """Base class for Feature Selector from scores."""

    # Abstract
    def get_feature_list(score_matrix: DataFrame,
                         **kwargs) -> Union[list[str], dict]:
        """A function to return top features from given scores of each feature for each class. 

        Args:
            score_matrix (DataFrame): Score of each feature across all classes
                                      [num_classes X num_features].

        Returns:
            list[str]: List of features.
        """
        return

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params."""
        return dict()


def build_selector(selector_config: dict) -> tuple[SelectorBase, dict]:
    """Builder object to get Selector, updated selector_config."""
    return build_object(scalr.feature.selector, selector_config)



"""This file generates gene recall curves for reference genes in provided models ranked genes"""

import json
from os import path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from scalr.analysis import AnalysisBase
from scalr.feature.selector import build_selector
from scalr.utils import EventLogger
from scalr.utils import read_data


class GeneRecallCurve(AnalysisBase):
    '''Class to generate gene recall curve.'''

    def __init__(self,
                 reference_genes_path: str,
                 ranked_genes_path_dict: dict = None,
                 top_K: int = 150,
                 plots_per_row: int = 3,
                 save_plots: bool = True,
                 features_selector: dict = None):
        '''Load required parameters for gene recall curve.

        Args:
            reference_genes_path: Reference genes csv path.
            ranked_genes_path: List of ranked genes csvs we want to check recall against. The list is used to support
                               plotting gene recall for multiple models at once.
            top_K: Top K genes to consider from ranked genes to check gene recall.
            plots_per_row: Number of categories to plot gene recall per row.
            save_plots: Flag indicating whether to store gene recall curves or not.
            feature_selector: Feature selection config containing aggregation strategy for `score_matrix`.
        '''
        self.ranked_genes_path_dict = ranked_genes_path_dict
        self.reference_genes_path = reference_genes_path
        self.top_K = top_K
        self.plots_per_row = plots_per_row
        self.save_plots = save_plots

        # Build feature selector if `score_matrix` is to be used to generate gene recall curve.
        if features_selector:
            self.selector, _ = build_selector(features_selector)

        self.event_logger = EventLogger('Gene Recall Curve analysis')

    def generate_analysis(self,
                          score_matrix: pd.DataFrame = None,
                          dirpath: str = '.',
                          **kwargs):
        '''This function calls function to generate gene recall after setting a few parameters.

        Args:
            score_matrix: Matrix that contains a score of each gene for each category.
            dirpath: Path to store gene recall curve if applicable.
        '''

        self.event_logger.heading2("Gene Recall Curve analysis")

        # Handle to ensure reference genes are provided.
        if not self.reference_genes_path:
            raise ValueError(
                'Reference genes are required for generating gene recall curves!'
            )
        reference_genes_df = read_data(self.reference_genes_path, index_col=0)

        # Extracting ranked genes.
        if not self.ranked_genes_path_dict:
            try:
                ranked_genes_df_dict = {
                    'pipeline_model':
                        pd.DataFrame.from_dict(
                            self.selector.get_feature_list(
                                score_matrix=score_matrix))
                }
            except:
                raise ValueError(
                    'There is some issue in `score_matrix` that is generated during the pipeline run. Please check!'
                )
        else:
            ranked_genes_df_dict = {}
            for model_name, genes_path in self.ranked_genes_path_dict.items():
                ranked_genes_df_dict[model_name] = read_data(genes_path,
                                                             index_col=0)

        # Generate gene recall curve.
        self.plot_gene_recall(ranked_genes_df_dict=ranked_genes_df_dict,
                              reference_genes_df=reference_genes_df,
                              dirpath=dirpath)

    def plot_gene_recall(self,
                         ranked_genes_df_dict: dict,
                         reference_genes_df: pd.DataFrame,
                         dirpath: str = '.',
                         title: str = ''):
        """This function plots & stores the gene recall curve for reference genes in provided ranked genes.

        It also stores the reference genes along with their ranks for each model in a json file for further
        analysis to the user.

        Args:
            ranked_genes_df_dict: Pipeline generated ranked genes dataframe.
            reference_genes_df: Reference genes dataframe.
            top_K: The top K-ranked genes in which reference genes are to be looked for.
            dirpath: Path to store gene recall plot and json.
            plot_type: Type of gene recall - per category or aggregated across all categories.
        """

        # Common categories across reference and ranked genes dataframes.
        common_categories = reference_genes_df.columns
        for model in ranked_genes_df_dict.values():
            common_categories = set(common_categories).intersection(
                model.columns)
        n_plots = len(common_categories)
        self.event_logger.info(
            f'\n{n_plots} categories matches between ranked genes & reference genes dataframes, namely: {common_categories}'
        )
        n_cols = min(self.plots_per_row, n_plots)
        n_rows = int(np.ceil(n_plots / n_cols))
        fig, axs = plt.subplots(n_rows,
                                n_cols,
                                figsize=(n_cols * self.plots_per_row,
                                         n_rows * self.plots_per_row),
                                squeeze=False)
        axs = axs.flatten()
        fig.suptitle(f'Recall of reference genes w.r.t ranked genes - {title}')

        # Plotting baseline gene recall curve.
        for i, category in enumerate(common_categories):
            ranked_genes = list(
                ranked_genes_df_dict.values())[0][category].values
            ref_genes = reference_genes_df[category].dropna().values
            k = self.top_K
            if k < len(ref_genes):
                k = len(ref_genes)

            if not len(ranked_genes) or not len(ref_genes):
                raise Exception(
                    'Ranked genes or ref genes list cannot be empty.')

            # Adjusting k if expected k > number of ranked genes.
            k = min(k, len(ranked_genes))

            # Building baseline curve.
            step = k // len(ref_genes)
            baseline = [i // step for i in range(1, 1 + k)]

            axs[i].plot(list(range(1, 1 + k)), baseline, label='baseline')
            axs[i].legend()

        # Plotting gene recall for defined ranked genes dataframes.
        gene_recall_dict = {}
        for df_name, ranked_genes_df in ranked_genes_df_dict.items():
            gene_recall_dict[df_name] = {}
            for i, category in enumerate(common_categories):
                ranked_genes = ranked_genes_df[category].values
                ref_genes = reference_genes_df[category].dropna().values
                k = self.top_K

                self.event_logger.info(f'\nCategory - {category}')
                # Removing reference genes that are not available in the list of the ranked genes.
                self.event_logger.info(
                    f'-- Number of reference genes provided : {len(ref_genes)}')
                ref_genes = list(set(ref_genes).intersection(ranked_genes))
                self.event_logger.info(
                    f'-- Number of reference genes found in ranked genes : {len(ref_genes)}'
                )

                if k < len(ref_genes):
                    self.event_logger.info(
                        f'-- top_K={k} should be greater than or equal to #reference_genes={len(ref_genes)}, setting'
                        f' top_K= #reference_genes({len(ref_genes)})')
                    k = len(ref_genes)

                if not len(ranked_genes) or not len(ref_genes):
                    raise Exception(
                        'Ranked genes or ref genes list cannot be empty.')

                # Adjusting k if expected k > number of ranked genes.
                if k > len(ranked_genes):
                    self.event_logger.info(
                        f'-- Setting k={len(ranked_genes)} as #ranked_genes < {k}'
                    )
                    k = len(ranked_genes)

                order_in_lit = {}
                points = []

                count = 0
                for rank, gene in enumerate(ranked_genes[:k]):
                    if gene in ref_genes:
                        count += 1
                        order_in_lit[rank] = gene
                    points.append(count)

                axs[i].plot(list(range(1, 1 + k)), points, label=df_name)
                axs[i].set_title(category)
                axs[i].set_xlabel('# Top-ranked genes (k)')
                axs[i].set_ylabel('# Reference genes')
                axs[i].legend()

                gene_recall_dict[df_name][category] = order_in_lit

        for j in range(i + 1, len(axs)):
            fig.delaxes(axs[j])

        plt.tight_layout()
        if self.save_plots:
            plt.savefig(path.join(dirpath, 'gene_recall_curve.svg'))
            with open(path.join(dirpath, 'gene_recall_curve_info.json'),
                      'w') as f:
                json.dump(gene_recall_dict, f, indent=6)
            self.event_logger.info(
                f'\nGene recall curves stored at path: `{path.join(dirpath, "gene_recall_curve.svg")}`'
            )
        else:
            plt.show()

        plt.close()

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for preprocess_config."""
        return dict(ranked_genes_path_dict={},
                    top_K=150,
                    plots_per_row=3,
                    save_plots=True,
                    features_selector={})



"""This file is an implementation of group splitter."""

from pandas import DataFrame
from sklearn.model_selection import GroupShuffleSplit

from scalr.data.split import StratifiedSplitter


class GroupSplitter(StratifiedSplitter):
    """Class for splitting data based on the provided group.

    Generate a stratified split of data into train, validation, and test
    sets. Stratification ensures samples have the same value for `stratify`
    column, can not belong to different sets.
    """

    def __init__(self, split_ratio: list[float], stratify: str):
        """Initialize splitter with required parameters.

        Args:
            split_ratio (list[float]): Ratio to split number of samples in.
            stratify (str): Column name to metadata the split upon in `obs`.
        """
        super().__init__(split_ratio)
        self.stratify = stratify

    def _split_data_with_stratification(
            self, metadata: DataFrame, target: str,
            test_ratio: float) -> tuple[list[int], list[int]]:
        """A function to split given metadata into a training and testing set.

        Args:
            metadata (DataFrame): Dataframe containing all samples to be split.
            target (str): Target for classification present in `obs`.
            test_ratio (float): Ratio of samples belonging to the test split.

        Returns:
            (list(int), list(int)): Two lists consisting of train and test indices.
        """
        splitter = GroupShuffleSplit(test_size=test_ratio,
                                     n_splits=1,
                                     random_state=42)

        train_inds, test_inds = next(
            splitter.split(metadata,
                           metadata[target],
                           groups=metadata[self.stratify]))

        return train_inds, test_inds

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for model_config."""
        return dict(split_ratio=[7, 1, 2], stratify='donor_id')



"""This is a test file for file_utils.py"""

import os
from os import path
import shutil

import numpy as np

from scalr.utils import generate_dummy_anndata
from scalr.utils import read_data
from scalr.utils import write_chunkwise_data
from scalr.utils import write_data


def test_write_chunkwise_data():
    """This function tests `write_chunkwise()`, `write_data()` & `read_data()` functions
    of file_utils."""
    os.makedirs('./tmp', exist_ok=True)

    # Generating dummy anndata.
    adata = generate_dummy_anndata(n_samples=25, n_features=5)

    # Path to write full data.
    fulldata_path = './tmp/fulldata.h5ad'
    write_data(adata, fulldata_path)

    # sample_chunksize to store full data in chunks.
    sample_chunksize = 5

    # Path to store chunked data.
    dirpath = './tmp/chunked_data/'

    # Writing fulldata in chunks.
    full_data = read_data(fulldata_path)
    write_chunkwise_data(full_data,
                         sample_chunksize=sample_chunksize,
                         dirpath=dirpath)

    # Iterating over stored chunked data to assert shape.
    observed_n_chunks = 0
    for i in range(len(os.listdir(dirpath))):
        if os.path.exists(path.join(dirpath, f'{i}.h5ad')):
            chunked_data = read_data(path.join(dirpath, f'{i}.h5ad'),
                                     backed='r')
            assert chunked_data.shape == (
                sample_chunksize, len(adata.var_names)
            ), f"There is some issue with chunk-{i}. Please check!"
            observed_n_chunks += 1
        else:
            break

    # Checking the number of chunks stored.
    expected_n_chunks = np.ceil(adata.shape[0] / sample_chunksize).astype(int)
    assert observed_n_chunks == expected_n_chunks, f"There is mismatch of observed_n_chunks - {observed_n_chunks} with expected_n_chunks - {expected_n_chunks}."

    shutil.rmtree('./tmp', ignore_errors=True)



"""This file contains functions related to data utility."""

from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from sklearn.preprocessing import OneHotEncoder
import torch


def get_one_hot_matrix(data: np.array):
    """This function returns a one-hot matrix of given labels.

    Args:
        data: Categorical data of dim 1D or 2D array.

    Returns:
        one-hot matrix.
    """
    if data.ndim == 1:
        data = data.reshape(-1, 1)
    ohe = OneHotEncoder().fit(data)
    one_hot_matrix = ohe.transform(data).toarray()

    return one_hot_matrix


def get_random_samples(
    data: Union[AnnData, AnnCollection],
    n_random_samples: int,
) -> torch.tensor:
    """This function returns random N samples from given data.

    Args:
        data: AnnData or AnnCollection object.
        n_random_samples: number of random samples to extract from the data.

    Returns:
        Chosen random samples tensor.
    """

    random_indices = np.random.randint(0, data.shape[0], n_random_samples)
    random_background_data = data[random_indices].X

    if not isinstance(random_background_data, np.ndarray):
        random_background_data = random_background_data.A

    random_background_data = torch.as_tensor(random_background_data,
                                             dtype=torch.float32)

    return random_background_data


def generate_dummy_anndata(n_samples, n_features, target_name='celltype'):
    """This function returns anndata object of shape (n_samples, n_features).
    
    It generates random values for target, batch & env from below mentioned choices.
    If you require more columns, you can add them in the below adata.obs without editing
    already existing columns.

    Args:
        n_samples: Number of samples in anndata.
        n_features: Number of features in anndata.
        target_name: Any preferred target name. Default is `celltype`.

    Returns:
        Anndata object.
    """

    # Setting seed for reproducibility.
    np.random.seed(0)

    # Creating anndata object.
    adata = AnnData(X=np.random.rand(n_samples, n_features))
    adata.obs = pd.DataFrame.from_dict({
        target_name: np.random.choice(['B', 'C', 'DC', 'T'], size=n_samples),
        'batch': np.random.choice(['batch1', 'batch2'], size=n_samples),
        'env': np.random.choice(['env1', 'env2', 'env3'], size=n_samples)
    })
    adata.obs.index = adata.obs.index.astype(str)

    return adata


def generate_dummy_dge_anndata(n_donors: int = 5,
                               cell_type_list: list[str] = [
                                   'B_cell', 'T_cell', 'DC'
                               ],
                               cell_replicate: int = 2,
                               n_vars: int = 10) -> AnnData:
    """This function returns anndata object for DGE analysis
    with shape (n_donors*len(cell_type_list)*cell_replicate, n_vars).

    It generates obs with random donors with a fixed clinical condition (disease_x or normal).
    Includes all the cell types in `cell_type_list` with number of `cell_replicate` for each donor.
    It generates a csr(Compressed Sparse Row) matrix with random gene expression values.
    It generates var with random gene name as `var.index` of length `n_vars`.

    Args:
        n_donors: Number of donors or subjects in `anndata.obs`.
        cell_type_list: List of different cell types to include.
        cell_replicate: Number of cell replicates per cell type.
        n_vars: Number of genes to include in `anndata.var`.

    Returns:
        Anndata object.
    """

    # Setting seed for reproducibility.
    np.random.seed(0)

    donor_list = [f'D_{i}' for i in range(1, n_donors + 1)]
    condition_array = np.random.choice(['disease_x', 'normal'],
                                       size=n_donors,
                                       replace=True)

    # Creating obs
    obs_data = []
    for donor, condition in zip(donor_list, condition_array):
        obs_data.extend([{
            'donor_id': donor,
            'cell_type': cell_type,
            'disease': condition
        } for i in range(cell_replicate) for cell_type in cell_type_list])
    obs = pd.DataFrame(obs_data)
    n_obs = obs.shape[0]
    obs.index = obs.index.astype(str)

    # Random geneexpression matrix
    X = csr_matrix(np.random.rand(n_obs, n_vars))

    # Creating var
    var = pd.DataFrame({
        'gene_id': [f'gid_{i}' for i in range(1, n_vars + 1)],
        'gene_name': [f'gene_{i}' for i in range(1, n_vars + 1)]
    }).set_index('gene_name')
    var.index = var.index.astype(str)

    # Creating AnnData object
    adata = AnnData(X=X, obs=obs, var=var)

    return adata



"""This file returns top K promoter features per class."""

from pandas import DataFrame

from scalr.feature.selector import SelectorBase


class ClasswisePromoters(SelectorBase):
    """Class for class-wise promoter feature selector strategy.
    
    Classwise scorer returns a dict for each class, containing the top
    positive scored genes.
    """

    def __init__(self, k: int = 1e6):
        """Initialize required parameters for the selector."""
        self.k = k

    def get_feature_list(self, score_matrix: DataFrame):
        """A function to return top features per class using score matrix
        and selector strategy.

        Args:
            score_matrix (DataFrame): Score of each feature across all classes
                                      [num_classes X num_features].

        Returns:
            list[str]: List of top k features.
        """
        classwise_promoters = dict()
        n_cls = len(score_matrix)

        for i in range(n_cls):
            for i in range(n_cls):
                classwise_promoters[score_matrix.index[i]] = score_matrix.iloc[
                    i, :].sort_values(ascending=False).reset_index(
                    )['index'][:self.k].tolist()

        return classwise_promoters

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for preprocess_config."""
        return dict(k=int(1e6))



"""This file is a base class for the analysis module."""

import os
from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
from torch import nn
from torch.utils.data import DataLoader

import scalr
from scalr.utils import build_object


class AnalysisBase:
    """A base class for downstream analysis tasks.
    
    This class provides common attributes and methods for all the analysis tasks.
    It is intended to be subclassed to create task-specific analysis.
    """

    def __init__(self):
        pass

    # Abstract
    def generate_analysis(self, model: nn.Module,
                          test_data: Union[AnnData, AnnCollection],
                          test_dl: DataLoader, dirpath: str, **kwargs):
        """A function to generate analysis, should be overridden by all subclasses.

        Args:
            model (nn.Module): final trained model.
            test_data (Union[AnnData, AnnCollection]): test data to run analysis on.
            test_dl (DataLoader): DataLoader object to prepare inputs for the model.
            dirpath (str): dirpath to store analysis.
            **kwargs: contains all previous analysis done to be used later.
        """
        pass

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for analysis_config."""
        return dict()


def build_analyser(analysis_config: dict) -> tuple[AnalysisBase, dict]:
    """Builder object to get analyser, updated analyser_config."""
    return build_object(scalr.analysis, analysis_config)



"""This file is a base class for a model trainer."""

from copy import deepcopy
import os
from os import path
from time import time

import torch
from torch import nn
from torch.nn import Module
from torch.optim import Optimizer
from torch.utils.data import DataLoader

from scalr.nn.callbacks import CallbackExecutor
from scalr.utils import EventLogger


class TrainerBase:
    """ Class for a model trainer. It trains and validates a model."""

    def __init__(self,
                 model: Module,
                 opt: Optimizer,
                 loss_fn: Module,
                 callbacks: CallbackExecutor,
                 device: str = 'cpu'):
        """Initialize required parameters for a model trainer.

        Args:
            model (Module): Model to train.
            opt (Optimizer): Optimizer used for learning.
            loss_fn (Module): Loss function used for training.
            callbacks (CallbackExecutor): Callback executor object to carry out callbacks.
            device (str, optional): Device to train the data on (cuda/cpu). Defaults to 'cpu'.
        """
        self.event_logger = EventLogger('ModelTrainer')

        self.model = model
        self.opt = opt
        self.loss_fn = loss_fn
        self.callbacks = callbacks
        self.device = device

    def train_one_epoch(self, dl: DataLoader) -> tuple[float, float]:
        """This function trains the model for one epoch.

        Args:
            dl: Training dataloader.

        Returns:
            Train Loss, Train Accuracy.
        """
        self.model.train()
        total_loss = 0
        hits = 0
        total_samples = 0
        for batch in dl:
            x, y = [example.to(self.device) for example in batch[:-1]
                   ], batch[-1].to(self.device)

            out = self.model(*x)['cls_output']
            loss = self.loss_fn(out, y)

            #training
            self.opt.zero_grad()
            loss.backward()
            self.opt.step()

            #logging
            total_loss += loss.item() * x[0].size(0)
            total_samples += x[0].size(0)
            hits += (torch.argmax(out, dim=1) == y).sum().item()

        total_loss /= total_samples
        accuracy = hits / total_samples
        return total_loss, accuracy

    def validation(self, dl: DataLoader) -> tuple[float, float]:
        """This function performs validation of the data.

        Args:
            dl: Validation dataloader.

        Returns:
            Validation Loss, Validation Accuracy.
        """
        self.model.eval()
        total_loss = 0
        hits = 0
        total_samples = 0
        for batch in dl:
            with torch.no_grad():
                x, y = [example.to(self.device) for example in batch[:-1]
                       ], batch[-1].to(self.device)
                out = self.model(*x)['cls_output']
                loss = self.loss_fn(out, y)

            #logging
            hits += (torch.argmax(out, dim=1) == y).sum().item()
            total_loss += loss.item() * x[0].size(0)
            total_samples += x[0].size(0)

        total_loss /= total_samples
        accuracy = hits / total_samples

        return total_loss, accuracy

    def train(self, epochs: int, train_dl: DataLoader, val_dl: DataLoader):
        """This function trains the model, and executes callbacks.

        Args:
            epochs: Max number of epochs to train model on.
            train_dl: Training dataloader.
            val_dl: Validation dataloader.
        """
        best_val_acc = 0
        best_model = deepcopy(self.model)

        for epoch in range(epochs):
            ep_start = time()
            self.event_logger.info(f'Epoch {epoch+1}:')
            train_loss, train_acc = self.train_one_epoch(train_dl)
            self.event_logger.info(
                f'Training Loss: {train_loss} || Training Accuracy: {train_acc}'
            )
            val_loss, val_acc = self.validation(val_dl)
            self.event_logger.info(
                f'Validation Loss: {val_loss} || Validation Accuracy: {val_acc}'
            )
            ep_end = time()
            self.event_logger.info(f'Time: {ep_end-ep_start}\n')

            if val_acc > best_val_acc:
                best_val_acc = val_acc
                best_model = deepcopy(self.model)

            if self.callbacks.execute(model_state_dict=self.model.state_dict(),
                                      opt_state_dict=self.opt.state_dict(),
                                      train_loss=train_loss,
                                      train_acc=train_acc,
                                      val_loss=val_loss,
                                      val_acc=val_acc):
                break

        return best_model



"""This file is the implementation of simpledataloader."""

from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
import torch
from torch import Tensor
from torch.nn.functional import pad

from scalr.nn.dataloader import DataLoaderBase


class SimpleDataLoader(DataLoaderBase):
    """Class for simple dataloader.
    
    Simple DataLoader converts all adata values to inputs, and target columns in metadata
    to output labels.

    Returns:
        PyTorch DataLoader object with (X: Tensor [batch_size, features], y: Tensor [batch_size, ]).
    """

    def __init__(self,
                 batch_size: int,
                 target: str,
                 mappings: dict,
                 padding: int = None):
        """
        Args:
            batch_size (int): Number of samples to be loaded in each batch.
            target (str): Corresponding metadata name to be treated as training
                          objective in classification. Must be present as a column_name in `adata.obs`.
            mappings (dict): Mapping the target name to respective ids.
            padding (int): Padding size in case of #features < model input size.
        """
        super().__init__(batch_size, target, mappings)
        self.padding = padding

    def collate_fn(
        self,
        adata_batch: Union[AnnData, AnnCollection],
    ) -> tuple[Tensor, Tensor]:
        """Given an input anndata of batch_size, the collate function creates inputs and outputs.

        Args:
            adata_batch (Union[AnnData, AnnCollection]): Anndata view object with batch_size samples.

        Returns:
            Tuple(x, y): Input to model, output from data.
        """

        x = adata_batch.X.float()
        # Handle the case when observed #features are less than expected #features by the model.
        # Features(0s) are padded after actual features in that case to make it consistent for model training.
        if self.padding and x.shape[1] < self.padding:
            x = pad(x, (0, self.padding - x.shape[1]), 'constant', 0.0)
        y = self.get_targets_ids_from_mappings(adata_batch)[0]

        return x, y

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for model_config."""
        return dict(batch_size=1, target=None, mappings=dict())



"""This file generates accuracy, classification report and stores it."""

from os import path

from pandas import DataFrame
from sklearn.metrics import accuracy_score
from sklearn.metrics import classification_report
import torch
from torch import nn
from torch.utils.data import DataLoader

from scalr.utils import EventLogger
from scalr.utils import write_data


def get_accuracy(test_labels: list[int], pred_labels: list[int]) -> float:
    """A function to get accuracy for the predicted labels.

    Args:
        test_labels (list[int]): True labels from the test set.
        pred_labels (list[int]): Predicted labels from the trained model.

    Returns:
        float: accuracy score
    """
    event_logger = EventLogger('Accuracy')
    accuracy = accuracy_score(test_labels, pred_labels)
    event_logger.info(f'Accuracy: {accuracy}')
    return accuracy


def generate_and_save_classification_report(test_labels: list[int],
                                            pred_labels: list[int],
                                            dirpath: str,
                                            mapping: dict = None) -> DataFrame:
    """A function to generate a classificaton report from the actual and predicted data
    and store at `dirpath`.

    Args:
        test_labels: True labels from the test set.
        pred_labels: Predicted labels from the trained model.
        dirpath: Path to store classification_report.
        mapping[optional]: Mapping of label_id to true label_names (id2label).

    Returns:
        A Pandas DataFrame with the classification report.
    """
    event_logger = EventLogger('ClassReport')

    if mapping:
        test_labels = [mapping[x] for x in test_labels]
        pred_labels = [mapping[x] for x in pred_labels]

    report = DataFrame(
        classification_report(test_labels, pred_labels,
                              output_dict=True)).transpose()
    event_logger.info('\nClassification Report:')
    event_logger.info(report)
    write_data(report, path.join(dirpath, 'classification_report.csv'))

    return report



"""This file contains implementation for model training on feature subsets."""

from copy import deepcopy
import os
from os import path
from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
from torch import nn

from scalr.model_training_pipeline import ModelTrainingPipeline
from scalr.utils import EventLogger
from scalr.utils import FlowLogger


class FeatureSubsetting:
    """Class for FeatureSubsetting.

    It trains a model for each subsetted datasets, each
    containing `feature_subsetsize` genes as features.
    """

    def __init__(self,
                 feature_subsetsize: int,
                 chunk_model_config: dict,
                 chunk_model_train_config: dict,
                 train_data: Union[AnnData, AnnCollection],
                 val_data: Union[AnnData, AnnCollection],
                 target: str,
                 mappings: dict,
                 dirpath: str = None,
                 device: str = 'cpu'):
        """Initialize required parameters for feature subset training.

        Args:
            feature_subsetsize (int): Number of features in one subset.
            chunk_model_config (dict): Chunked model config.
            chunk_model_train_config (dict): Chunked model training config.
            train_data (Union[AnnData, AnnCollection]): Train dataset.
            val_data (Union[AnnData, AnnCollection]): Validation dataset.
            target (str): Target to train model.
            mappings (dict): mapping of target to labels.
            dirpath (str, optional): Dirpath to store chunked model weights. Defaults to None.
            device (str, optional): Device to train models on. Defaults to 'cpu'.
        """
        self.event_logger = EventLogger('FeatureSubsetting')

        self.feature_subsetsize = feature_subsetsize
        self.chunk_model_config = chunk_model_config
        self.chunk_model_train_config = chunk_model_train_config
        self.train_data = train_data
        self.val_data = val_data
        self.target = target
        self.mappings = mappings
        self.dirpath = dirpath
        self.device = device

    def train_chunked_models(self) -> list[nn.Module]:
        """Trains a model for each subset data.

        Returns:
            list[nn.Module]: List of models for each subset.
        """
        self.event_logger.info('Feature subset models training')
        models = []
        chunked_models_dirpath = path.join(self.dirpath, 'chunked_models')
        os.makedirs(chunked_models_dirpath, exist_ok=True)

        i = 0
        for start in range(0, len(self.train_data.var_names),
                           self.feature_subsetsize):
            self.event_logger.info(f'\nChunk {i}')
            chunk_dirpath = path.join(chunked_models_dirpath, str(i))
            os.makedirs(chunk_dirpath, exist_ok=True)
            i += 1

            train_features_subset = self.train_data[:, start:start +
                                                    self.feature_subsetsize]
            val_features_subset = self.val_data[:, start:start +
                                                self.feature_subsetsize]

            chunk_model_config = deepcopy(self.chunk_model_config)

            model_trainer = ModelTrainingPipeline(chunk_model_config,
                                                  self.chunk_model_train_config,
                                                  chunk_dirpath, self.device)

            model_trainer.set_data_and_targets(train_features_subset,
                                               val_features_subset, self.target,
                                               self.mappings)
            model_trainer.build_model_training_artifacts()
            best_model = model_trainer.train()

            self.chunk_model_config, self.chunk_model_train_config = model_trainer.get_updated_config(
            )

            models.append(best_model)

        return models

    def get_updated_configs(self):
        """Returns updated configs."""
        return self.chunk_model_config, self.chunk_model_train_config



"""This file performs standard scaler normalization on the data."""

from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
import numpy as np

from scalr.data.preprocess import PreprocessorBase
from scalr.utils import EventLogger


class StandardScaler(PreprocessorBase):
    """Class for Standard Normalization"""

    def __init__(self, with_mean: bool = True, with_std: bool = True):
        """Initialize parameters for standard scaler normalization.

        Args:
            with_mean: Mean for standard scaling.
            with_std: Standard deviation for standard scaling.
        """

        self.with_mean = with_mean
        self.with_std = with_std

        # Parameters for standard scaler.
        self.train_mean = None
        self.train_std = None

        self.event_logger = EventLogger('Standard Scaler Normalization')

    def transform(self, data: np.ndarray) -> np.ndarray:
        """A function to transform provided input data.

        Args:
            data (np.ndarray): raw data

        Returns:
            np.ndarray: processed data
        """
        self.event_logger.info(
            '\Transforming data using standard scaler object')

        if not self.with_mean:
            train_mean = np.zeros((1, data.shape[1]))
        else:
            train_mean = self.train_mean
        return (data - train_mean) / self.train_std

    def fit(self, data: Union[AnnData, AnnCollection],
            sample_chunksize: int) -> None:
        """This function calculate parameters for standard scaler object from the train data.
        
        Args:
            data: Data to calculate the required parameters of.
            sample_chunksize: Chunks of data that can be loaded into memory at once.
        
        """

        self.event_logger.info('\n\nStarting standardscaler normalization')
        self.event_logger.info('\nFitting standard scaler object on train data')

        self.calculate_mean(data, sample_chunksize)
        self.calculate_std(data, sample_chunksize)

    def calculate_mean(self, data: Union[AnnData, AnnCollection],
                       sample_chunksize: int) -> None:
        """Function to calculate mean for each feature in the train data
        
        Args:
            data: Data to calculate the mean of.
            sample_chunksize: Chunks of data that can be loaded into memory at once.
            
        Returns:
            Nothing, stores mean per feature of the train data.
            """

        self.event_logger.info('Calculating mean of data...')
        train_sum = np.zeros(data.shape[1]).reshape(1, -1)

        # Iterate through batches of data to get mean statistics
        for i in range(int(np.ceil(data.shape[0] / sample_chunksize))):
            train_sum += data[i * sample_chunksize:i * sample_chunksize +
                              sample_chunksize].X.sum(axis=0)
        self.train_mean = train_sum / data.shape[0]

        if not self.with_mean:
            self.event_logger.info(
                '`train_mean` will be set to zero during `transform()`, as `with_mean` is set to False!'
            )

    def calculate_std(self, data: Union[AnnData, AnnCollection],
                      sample_chunksize: int) -> None:
        """A function to calculate standard deviation for each feature in the train data.
        
        Args:
            data: Data to calculate the standard deviation of
            sample_chunksize: Chunks of data that can be loaded into memory at once.
            
        Returns:
            Nothing, stores standard deviation per feature of the train data.
            """

        # Getting standard deviation of entire train data per feature.
        if self.with_std:
            self.event_logger.info('Calculating standard deviation of data...')
            self.train_std = np.zeros(data.shape[1]).reshape(1, -1)
            # Iterate through batches of data to get std statistics
            for i in range(int(np.ceil(data.shape[0] / sample_chunksize))):
                self.train_std += np.sum(np.power(
                    data[i * sample_chunksize:i * sample_chunksize +
                         sample_chunksize].X - self.train_mean, 2),
                                         axis=0)
            self.train_std /= data.shape[0]
            self.train_std = np.sqrt(self.train_std)

            # Handling cases where standard deviation of feature is 0, replace it with 1.
            self.train_std[self.train_std == 0] = 1
        else:
            # If `with_std` is False, set train_std to 1.
            self.event_logger.info(
                'Setting `train_std` to be 1, as `with_std` is set to False!')
            self.train_std = np.ones((1, data.shape[1]))

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for preprocess_config."""
        return dict(with_mean=True, with_std=True)



"""This file is a base class for dataloader."""

from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
from anndata.experimental import AnnLoader
import torch
from torch import Tensor

import scalr
from scalr.utils import build_object


class DataLoaderBase:

    def __init__(
        self,
        batch_size: int = 1,
        target: Union[str, list[str]] = None,
        mappings: dict = None,
    ):
        """Initilize required parameters for dataloader.

        Args:
            batch_size (int, optional): _description_. Defaults to 1.
            target ([str, list[str]]): List of target. Defaults to None.
            mappings (dict): List of label mappings of each target to. Defaults to None.
        """
        self.batch_size = batch_size
        self.target = target
        self.mappings = mappings

    # Abstract
    def collate_fn(self, batch):
        """Collate function for dataloader. Should be implemented in child classes.
        
        Given an input anndata of batch_size, the collate function creates inputs and outputs.
        It can also be used to perform batch-wise operations.
        """
        pass

    def get_targets_ids_from_mappings(
            self, adata: Union[AnnData, AnnCollection]) -> list[Tensor]:
        """Helper function to generate target ids from label mappings.

        Args:
            adata (Union[AnnData, AnnCollection]): Anndata object containing targets in `obs`.
        """
        target_ids = []
        if isinstance(self.target, str):
            targets = [self.target]
        else:
            targets = self.target

        for target in targets:
            target_mappings = self.mappings[target]['label2id']
            target_ids.append(
                torch.as_tensor(
                    adata.obs[self.target].astype('category').cat.
                    rename_categories(target_mappings).astype('int64').values))

        return target_ids

    def __call__(self, adata):
        """Returns a Torch DataLoader object."""
        return AnnLoader(adata,
                         batch_size=self.batch_size,
                         collate_fn=lambda batch: self.collate_fn(batch))

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for model_config."""
        return dict()


def build_dataloader(dataloader_config: dict,
                     adata: Union[AnnData, AnnCollection], target: str,
                     mappings: dict) -> tuple[DataLoaderBase, dict]:
    """Builder object to get DataLoader, updated dataloader_config.

    Args:
        dataloader_config (dict): Config to build dataloader.
        adata (Union[AnnData, AnnCollection]): Data to load.
        target (str): Target column in data.obs.
        mappings (dict): Mappings of column labels to ids.
    """
    if not dataloader_config.get('name'):
        raise ValueError('DataLoader name is required!')

    dataloader_config['params'] = dataloader_config.get('params',
                                                        dict(batch_size=1))
    dataloader_config['params']['target'] = target
    dataloader_config['params']['mappings'] = mappings

    dataloader_object, dataloader_config = build_object(scalr.nn.dataloader,
                                                        dataloader_config)
    dataloader_config['params'].pop('target')
    dataloader_config['params'].pop('mappings')

    return dataloader_object(adata), dataloader_config



"""This is a test file for Sample-wise normalization."""

import scanpy as sc

from scalr.data.preprocess import sample_norm
from scalr.utils import generate_dummy_anndata


def test_transform():
    '''This function tests the transform function of Sample-wise normalization.

    There is no fit() involved in Sample-wise normalization.
    '''

    # Creating an annadata object.
    adata = generate_dummy_anndata(n_samples=100, n_features=25)

    # Sample-wise norm required parameter.
    target_sum = 5

    # scalr sample-wise normalization.
    scalr_sample_norm = sample_norm.SampleNorm(scaling_factor=target_sum)
    # No need to fit() for sample-norm normalization
    scalr_scaled_data = scalr_sample_norm.transform(adata.X)

    # scanpy sample-wise normalization.
    scanpy_scaled_data = sc.pp.normalize_total(adata,
                                               target_sum=target_sum,
                                               inplace=False)['X']

    # asserts to check transformed data having errors less than 1e-15 compared to scanpy's transformed data.
    assert sum(
        abs(scanpy_scaled_data.flatten() -
            scalr_scaled_data.flatten()).flatten() < 1e-15
    ) == scalr_scaled_data.flatten().shape[
        0], "The sample norm is incorrectly transforming data, please debug code."



"""This file contains functions related to miscellaneous utilities."""

import os
import random

import numpy as np
import torch


def set_seed(seed: int):
    """A function to set seed for reproducibility."""
    os.environ['PYTHONHASHSEED'] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def overwrite_default(user_config: dict, default_config: dict) -> dict:
    """The function recursively overwrites information from user_config
    onto the default_config.
    """

    for key in user_config.keys():
        if key not in default_config.keys() or not isinstance(
                user_config[key], dict):
            default_config[key] = user_config[key]
        else:
            default_config[key] = overwrite_default(user_config[key],
                                                    default_config[key])

    return default_config


def build_object(module, config: dict):
    """A builder function to build an object from its config.

    Args:
        module: Module containing the class.
        config: Contains the name of the class and params to initialize the object.

    Returns: Object, updated config.
    """
    name = config.get('name')
    if not name:
        raise ValueError('class name not provided!')

    params = config.get('params', dict())
    default_params = getattr(module, name).get_default_params()
    params = overwrite_default(params, default_params)
    final_config = dict(name=name, params=params)

    return getattr(module, name)(**params), final_config



"""This file contains the implementation of feature subsetting, model training followed by top feature extraction."""

from copy import deepcopy
import os
from os import path
from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
import numpy as np
import pandas as pd
from torch import nn

from scalr.feature import FeatureSubsetting
from scalr.feature.scoring import build_scorer
from scalr.feature.selector import build_selector
from scalr.utils import FlowLogger
from scalr.utils import load_train_val_data_from_config
from scalr.utils import read_data
from scalr.utils import write_chunkwise_data
from scalr.utils import write_data


class FeatureExtractionPipeline:

    def __init__(self, feature_selection_config, dirpath, device):
        """Initialize required parameters for feature selection.

        Feature extraction is done in 4 steps:
        1. Model(s) training on chunked/all features
        2. Class X Feature scoring
        3. Top features extraction
        4. Feature subset data writing

        Args:
            feature_selection_config: Feature selection config.
            dirpath: Path to load data from.
        """
        self.flow_logger = FlowLogger('FeatureExtraction')

        self.feature_selection_config = deepcopy(feature_selection_config)
        self.device = device

        self.dirpath = dirpath
        os.makedirs(dirpath, exist_ok=True)

    def load_data_and_targets_from_config(self, data_config: dict):
        """A function to load data and targets from data config.

        Args:
            data_config: Data config.
        """
        self.train_data, self.val_data = load_train_val_data_from_config(
            data_config)
        self.target = data_config.get('target')
        self.mappings = read_data(data_config['label_mappings'])

    def set_data_and_targets(self, train_data: Union[AnnData, AnnCollection],
                             val_data: Union[AnnData, AnnCollection],
                             target: Union[str, list[str]], mappings: dict):
        """A function to set data when you don't use data directly from config,
        but rather by other sources like feature subsetting, etc.

        Args:
            train_data (Union[AnnData, AnnCollection]): Training data.
            val_data (Union[AnnData, AnnCollection]): Validation data.
            target (Union[str, list[str]]): Target columns name(s).
            mappings (dict): Mapping of a column value to ids
                            eg. mappings[column_name][label2id] = {A: 1, B:2, ...}.
        """
        self.train_data = train_data
        self.val_data = val_data
        self.target = target
        self.mappings = mappings

    def feature_subsetted_model_training(self) -> list[nn.Module]:
        """This function train models on subsetted data containing `feature_subsetsize` genes."""

        self.flow_logger.info('Feature subset models training')

        self.feature_subsetsize = self.feature_selection_config.get(
            'feature_subsetsize', len(self.val_data.var_names))

        chunk_model_config = self.feature_selection_config.get('model')
        chunk_model_train_config = self.feature_selection_config.get(
            'model_train_config')

        chunked_features_model_trainer = FeatureSubsetting(
            self.feature_subsetsize, chunk_model_config,
            chunk_model_train_config, self.train_data, self.val_data,
            self.target, self.mappings, self.dirpath, self.device)

        self.chunked_models = chunked_features_model_trainer.train_chunked_models(
        )
        chunk_model_config, chunk_model_train_config = chunked_features_model_trainer.get_updated_configs(
        )
        self.feature_selection_config['model'] = chunk_model_config
        self.feature_selection_config[
            'model_train_config'] = chunk_model_train_config

        return self.chunked_models

    def set_model(self, models: list[nn.Module]):
        """A function to set the trained model for downstream feature tasks."""
        self.chunked_models = models

    def feature_scoring(self) -> pd.DataFrame:
        """A function to generate scores of each feature for each class using a scorer
        and chunked models.
        """
        self.flow_logger.info('Feature scoring')

        scorer, scorer_config = build_scorer(
            deepcopy(self.feature_selection_config.get('scoring_config')))
        self.feature_selection_config['scoring_config'] = scorer_config

        all_scores = []
        if not getattr(self, 'feature_subsetsize', None):
            self.feature_subsetsize = self.train_data.shape[1]

        for i, (model) in enumerate(self.chunked_models):
            subset_train_data = self.train_data[:, i *
                                                self.feature_subsetsize:(i +
                                                                         1) *
                                                self.feature_subsetsize]
            subset_val_data = self.val_data[:, i *
                                            self.feature_subsetsize:(i + 1) *
                                            self.feature_subsetsize]
            score = scorer.generate_scores(model, subset_train_data,
                                           subset_val_data, self.target,
                                           self.mappings)

            all_scores.append(score[:self.feature_subsetsize])

        columns = self.train_data.var_names
        columns.name = "index"
        class_labels = self.mappings[self.target]['id2label']
        all_scores = np.concatenate(all_scores, axis=1)
        all_scores = all_scores[:, :len(columns)]

        self.score_matrix = pd.DataFrame(all_scores,
                                         columns=columns,
                                         index=class_labels)
        write_data(self.score_matrix, path.join(self.dirpath,
                                                'score_matrix.csv'))
        return self.score_matrix

    def set_score_matrix(self, score_matrix: pd.DataFrame):
        """A function to set score_matrix for feature extraction."""
        self.score_matrix = score_matrix

    def top_feature_extraction(self) -> Union[list[str], dict]:
        """A function to get top features using `Selector`."""

        self.flow_logger.info('Top features extraction')

        selector_config = self.feature_selection_config.get(
            'features_selector', dict(name='AbsMean'))
        selector, selector_config = build_selector(selector_config)
        self.feature_selection_config['features_selector'] = selector_config

        self.top_features = selector.get_feature_list(self.score_matrix)
        write_data(self.top_features,
                   path.join(self.dirpath, 'top_features.json'))

        return self.top_features

    def write_top_features_subset_data(self, data_config: dict) -> dict:
        """A function to write top features subset data onto disk
        and return updated data_config.

        Args:
            data_config: Data config.
        """

        self.flow_logger.info('Writing feature-subset data onto disk')

        datapath = data_config['train_val_test'].get('final_datapaths')
        feature_subset_datapath = path.join(self.dirpath, 'feature_subset_data')
        os.makedirs(feature_subset_datapath, exist_ok=True)

        test_data = read_data(path.join(datapath, 'test'))
        splits = {
            'train': self.train_data,
            'val': self.val_data,
            'test': test_data
        }

        for split, split_data in splits.items():

            split_feature_subset_datapath = path.join(feature_subset_datapath,
                                                      split)
            sample_chunksize = data_config.get('sample_chunksize')
            write_chunkwise_data(split_data,
                                 sample_chunksize,
                                 split_feature_subset_datapath,
                                 feature_inds=self.top_features)

        data_config['train_val_test'][
            'feature_subset_datapaths'] = feature_subset_datapath

        return data_config

    def get_updated_config(self) -> dict:
        """This function returns updated configs."""
        return self.feature_selection_config



from ._split import build_splitter
from ._split import SplitterBase
from .stratified_group_splitter import StratifiedGroupSplitter

# GroupSplitter inherit from StratifiedSplitter.
from .stratified_splitter import StratifiedSplitter    # isort:skip
from .group_splitter import GroupSplitter    # isort:skip
